/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2001
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (cd ..; . ./setit Epson.PDC; cd CUPS; ./GenerateOmniPPD driver libEpson_Stylus_C20.so jobprop "form=a4 media=MEDIA_SPECIAL_360")
 *
 */
#include <cstdio>
#include <cstdlib>

#include <cups/cups.h>
#include <cups/raster.h>

#include <glib.h>
#include <gmodule.h>

#include "Device.hpp"
#include "Omni.hpp"

#define TMM_TO_INCHES 25400.0
#define BUFSIZE1      64

void
printUsage (char *pszProgramName)
{
   std::cerr << "Usage: "
             << pszProgramName
             << " --driver printer_library_name"
             << " [-sproperties=\"properties\"]"
             << std::endl;
}

void
writePostscriptName (DeviceForm *pDF, char *pszBuffer)
{
   HardCopyCap *pHCC = pDF->getHardCopyCap ();
   int          iCx  = (int)(72.0 * (float)pHCC->getCx () / TMM_TO_INCHES);
   int          iCy  = (int)(72.0 * (float)pHCC->getCy () / TMM_TO_INCHES);

   switch (pDF->getID ())
   {
   case DeviceForm::FORM_LEDGER:                strcpy (pszBuffer, "FORM_LEDGER"); break;
   case DeviceForm::FORM_LEGAL:                 strcpy (pszBuffer, "FORM_LEGAL"); break;
   case DeviceForm::FORM_LEGAL_EXTRA:           strcpy (pszBuffer, "FORM_LEGAL_EXTRA"); break;
   case DeviceForm::FORM_LETTER:                strcpy (pszBuffer, "FORM_LETTER"); break;
   case DeviceForm::FORM_LETTER_PLUS:           strcpy (pszBuffer, "FORM_LETTER_PLUS"); break;
   case DeviceForm::FORM_LETTER_EXTRA:          strcpy (pszBuffer, "FORM_LETTER_EXTRA"); break;
   case DeviceForm::FORM_HALF_LETTER:           strcpy (pszBuffer, "FORM_HALF_LETTER"); break;
   case DeviceForm::FORM_EXECUTIVE:             strcpy (pszBuffer, "FORM_EXECUTIVE"); break;
   case DeviceForm::FORM_STATEMENT:             strcpy (pszBuffer, "FORM_STATEMENT"); break;
   case DeviceForm::FORM_TABLOID:               strcpy (pszBuffer, "FORM_TABLOID"); break;
   case DeviceForm::FORM_UNIVERSAL:             strcpy (pszBuffer, "FORM_UNIVERSAL"); break;
   case DeviceForm::FORM_FOOLSCAP:              strcpy (pszBuffer, "FORM_FOOLSCAP"); break;
   case DeviceForm::FORM_QUARTO:                strcpy (pszBuffer, "FORM_QUARTO"); break;
   case DeviceForm::FORM_FOLIO:                 strcpy (pszBuffer, "FORM_FOLIO"); break;
   case DeviceForm::FORM_FOLIO_SP:              strcpy (pszBuffer, "FORM_FOLIO_SP"); break;
   case DeviceForm::FORM_ROC_8K:                strcpy (pszBuffer, "FORM_ROC_8K"); break;
   case DeviceForm::FORM_ROC_16K:               strcpy (pszBuffer, "FORM_ROC_16K"); break;
   case DeviceForm::FORM_GOVERNMENT_LETTER:     strcpy (pszBuffer, "FORM_GOVERNMENT_LETTER"); break;
   case DeviceForm::FORM_GOVERNMENT_LEGAL:      strcpy (pszBuffer, "FORM_GOVERNMENT_LEGAL"); break;
   case DeviceForm::FORM_EDP:                   strcpy (pszBuffer, "FORM_EDP"); break;
   case DeviceForm::FORM_EUROPEAN_EDP:          strcpy (pszBuffer, "FORM_EUROPEAN_EDP"); break;
   case DeviceForm::FORM_PRC_16K:               strcpy (pszBuffer, "FORM_PRC_16K"); break;
   case DeviceForm::FORM_PRC_32K:               strcpy (pszBuffer, "FORM_PRC_32K"); break;
   case DeviceForm::FORM_DAI_PA_KAI:            strcpy (pszBuffer, "FORM_DAI_PA_KAI"); break;
///case DeviceForm::FORM_JURRO_KU_KAI:          strcpy (pszBuffer, "FORM_JURRO_KU_KAI"); break;
   case DeviceForm::FORM_PA_KAI:                strcpy (pszBuffer, "FORM_PA_KAI"); break;
   case DeviceForm::FORM_WIDE:                  strcpy (pszBuffer, "FORM_WIDE"); break;
   case DeviceForm::FORM_LETTER_WIDE:           strcpy (pszBuffer, "FORM_LETTER_WIDE"); break;
   case DeviceForm::FORM_FOOLSCAP_WIDE:         strcpy (pszBuffer, "FORM_FOOLSCAP_WIDE"); break;
   case DeviceForm::FORM_A3_WIDE:               strcpy (pszBuffer, "FORM_A3_WIDE"); break;
   case DeviceForm::FORM_A4_WIDE:               strcpy (pszBuffer, "FORM_A4_WIDE"); break;
   case DeviceForm::FORM_WIDE_FORMAT:           strcpy (pszBuffer, "FORM_WIDE_FORMAT"); break;
   case DeviceForm::FORM_10_X_11:               strcpy (pszBuffer, "FORM_10_X_11"); break;
   case DeviceForm::FORM_11_X_12:               strcpy (pszBuffer, "FORM_11_X_12"); break;
   case DeviceForm::FORM_11_X_15:               strcpy (pszBuffer, "FORM_11_X_15"); break;
   case DeviceForm::FORM_12_X_19:               strcpy (pszBuffer, "FORM_12_X_19"); break;
   case DeviceForm::FORM_15_X_11:               strcpy (pszBuffer, "FORM_15_X_11"); break;
   case DeviceForm::FORM_5_X_7:                 strcpy (pszBuffer, "FORM_5_X_7"); break;
   case DeviceForm::FORM_3_X_5_CARD:            strcpy (pszBuffer, "FORM_3_X_5_CARD"); break;
   case DeviceForm::FORM_4_X_6_CARD:            strcpy (pszBuffer, "FORM_4_X_6_CARD"); break;
   case DeviceForm::FORM_5_X_8_CARD:            strcpy (pszBuffer, "FORM_5_X_8_CARD"); break;
   case DeviceForm::FORM_8_X_10_CARD:           strcpy (pszBuffer, "FORM_8_X_10_CARD"); break;
   case DeviceForm::FORM_A6_CARD:               strcpy (pszBuffer, "FORM_A6_CARD"); break;
   case DeviceForm::FORM_CARD_148:              strcpy (pszBuffer, "FORM_CARD_148"); break;
   case DeviceForm::FORM_HAGAKI_CARD:           strcpy (pszBuffer, "FORM_HAGAKI_CARD"); break;
   case DeviceForm::FORM_OUFUKU_CARD:           strcpy (pszBuffer, "FORM_OUFUKU_CARD"); break;
   case DeviceForm::FORM_POSTCARD:              strcpy (pszBuffer, "FORM_POSTCARD"); break;
   case DeviceForm::FORM_INDEX_4_X_6_EXT:       strcpy (pszBuffer, "FORM_INDEX_4_X_6_EXT"); break;
   case DeviceForm::FORM_A2_ENVELOPE:           strcpy (pszBuffer, "FORM_A2_ENVELOPE"); break;
   case DeviceForm::FORM_C5_ENVELOPE:           strcpy (pszBuffer, "FORM_C5_ENVELOPE"); break;
   case DeviceForm::FORM_C6_ENVELOPE:           strcpy (pszBuffer, "FORM_C6_ENVELOPE"); break;
   case DeviceForm::FORM_C7_ENVELOPE:           strcpy (pszBuffer, "FORM_C7_ENVELOPE"); break;
   case DeviceForm::FORM_C9_ENVELOPE:           strcpy (pszBuffer, "FORM_C9_ENVELOPE"); break;
   case DeviceForm::FORM_C10_ENVELOPE:          strcpy (pszBuffer, "FORM_C10_ENVELOPE"); break;
   case DeviceForm::FORM_D5_ENVELOPE:           strcpy (pszBuffer, "FORM_D5_ENVELOPE"); break;
   case DeviceForm::FORM_DL_ENVELOPE:           strcpy (pszBuffer, "FORM_DL_ENVELOPE"); break;
   case DeviceForm::FORM_ENVELOPE_132_X_220:    strcpy (pszBuffer, "FORM_ENVELOPE_132_X_220"); break;
   case DeviceForm::FORM_ENVELOPE_6_X_9:        strcpy (pszBuffer, "FORM_ENVELOPE_6_X_9"); break;
   case DeviceForm::FORM_ENVELOPE_7_X_9:        strcpy (pszBuffer, "FORM_ENVELOPE_7_X_9"); break;
   case DeviceForm::FORM_ENVELOPE_6_1_2:        strcpy (pszBuffer, "FORM_ENVELOPE_6_1_2"); break;
   case DeviceForm::FORM_ENVELOPE_9_X_11:       strcpy (pszBuffer, "FORM_ENVELOPE_9_X_11"); break;
   case DeviceForm::FORM_ENVELOPE_10_X_13:      strcpy (pszBuffer, "FORM_ENVELOPE_10_X_13"); break;
   case DeviceForm::FORM_ENVELOPE_10_X_14:      strcpy (pszBuffer, "FORM_ENVELOPE_10_X_14"); break;
   case DeviceForm::FORM_ENVELOPE_10_X_15:      strcpy (pszBuffer, "FORM_ENVELOPE_10_X_15"); break;
   case DeviceForm::FORM_MONARCH_ENVELOPE:      strcpy (pszBuffer, "FORM_MONARCH_ENVELOPE"); break;
   case DeviceForm::FORM_PERSONAL_ENVELOPE:     strcpy (pszBuffer, "FORM_PERSONAL_ENVELOPE"); break;
   case DeviceForm::FORM_NUMBER_9_ENVELOPE:     strcpy (pszBuffer, "FORM_NUMBER_9_ENVELOPE"); break;
   case DeviceForm::FORM_NUMBER_10_ENVELOPE:    strcpy (pszBuffer, "FORM_NUMBER_10_ENVELOPE"); break;
   case DeviceForm::FORM_NUMBER_11_ENVELOPE:    strcpy (pszBuffer, "FORM_NUMBER_11_ENVELOPE"); break;
   case DeviceForm::FORM_NUMBER_12_ENVELOPE:    strcpy (pszBuffer, "FORM_NUMBER_12_ENVELOPE"); break;
   case DeviceForm::FORM_NUMBER_14_ENVELOPE:    strcpy (pszBuffer, "FORM_NUMBER_14_ENVELOPE"); break;
   case DeviceForm::FORM_ITALIAN_ENVELOPE:      strcpy (pszBuffer, "FORM_ITALIAN_ENVELOPE"); break;
   case DeviceForm::FORM_POSTFIX_ENVELOPE:      strcpy (pszBuffer, "FORM_POSTFIX_ENVELOPE"); break;
   case DeviceForm::FORM_INVITE_ENVELOPE:       strcpy (pszBuffer, "FORM_INVITE_ENVELOPE"); break;
   case DeviceForm::FORM_CHOU2_ENVELOPE:        strcpy (pszBuffer, "FORM_CHOU2_ENVELOPE"); break;
   case DeviceForm::FORM_CHOU3_ENVELOPE:        strcpy (pszBuffer, "FORM_CHOU3_ENVELOPE"); break;
   case DeviceForm::FORM_CHOU4_ENVELOPE:        strcpy (pszBuffer, "FORM_CHOU4_ENVELOPE"); break;
   case DeviceForm::FORM_KAHU_ENVELOPE:         strcpy (pszBuffer, "FORM_KAHU_ENVELOPE"); break;
   case DeviceForm::FORM_KAKU2_ENVELOPE:        strcpy (pszBuffer, "FORM_KAKU2_ENVELOPE"); break;
   case DeviceForm::FORM_YOU4_ENVELOPE:         strcpy (pszBuffer, "FORM_YOU4_ENVELOPE"); break;
   case DeviceForm::FORM_PRC1_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC1_ENVELOPE"); break;
   case DeviceForm::FORM_PRC2_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC2_ENVELOPE"); break;
   case DeviceForm::FORM_PRC3_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC3_ENVELOPE"); break;
   case DeviceForm::FORM_PRC4_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC4_ENVELOPE"); break;
   case DeviceForm::FORM_PRC5_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC5_ENVELOPE"); break;
   case DeviceForm::FORM_PRC6_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC6_ENVELOPE"); break;
   case DeviceForm::FORM_PRC7_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC7_ENVELOPE"); break;
   case DeviceForm::FORM_PRC8_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC8_ENVELOPE"); break;
   case DeviceForm::FORM_PRC9_ENVELOPE:         strcpy (pszBuffer, "FORM_PRC9_ENVELOPE"); break;
   case DeviceForm::FORM_PRC10_ENVELOPE:        strcpy (pszBuffer, "FORM_PRC10_ENVELOPE"); break;
   case DeviceForm::FORM_DISK_LABELS:           strcpy (pszBuffer, "FORM_DISK_LABELS"); break;
   case DeviceForm::FORM_EURO_LABELS:           strcpy (pszBuffer, "FORM_EURO_LABELS"); break;
   case DeviceForm::FORM_SHIPPING_LABELS:       strcpy (pszBuffer, "FORM_SHIPPING_LABELS"); break;
   case DeviceForm::FORM_STANDARD_LABELS_CLEAR: strcpy (pszBuffer, "FORM_STANDARD_LABELS_CLEAR"); break;
   case DeviceForm::FORM_STANDARD_LABELS_WHITE: strcpy (pszBuffer, "FORM_STANDARD_LABELS_WHITE"); break;
   case DeviceForm::FORM_2A0:                   strcpy (pszBuffer, "FORM_2A0"); break;
   case DeviceForm::FORM_4A0:                   strcpy (pszBuffer, "FORM_4A0"); break;
   case DeviceForm::FORM_A0:                    strcpy (pszBuffer, "FORM_A0"); break;
   case DeviceForm::FORM_A1:                    strcpy (pszBuffer, "FORM_A1"); break;
   case DeviceForm::FORM_A2:                    strcpy (pszBuffer, "FORM_A2"); break;
   case DeviceForm::FORM_A3:                    strcpy (pszBuffer, "FORM_A3"); break;
   case DeviceForm::FORM_A3_EXTRA:              strcpy (pszBuffer, "FORM_A3_EXTRA"); break;
   case DeviceForm::FORM_A4:                    strcpy (pszBuffer, "FORM_A4"); break;
   case DeviceForm::FORM_A4_EXTRA:              strcpy (pszBuffer, "FORM_A4_EXTRA"); break;
   case DeviceForm::FORM_A4_TAB:                strcpy (pszBuffer, "FORM_A4_TAB"); break;
   case DeviceForm::FORM_A5:                    strcpy (pszBuffer, "FORM_A5"); break;
   case DeviceForm::FORM_A5_EXTRA:              strcpy (pszBuffer, "FORM_A5_EXTRA"); break;
   case DeviceForm::FORM_A6:                    strcpy (pszBuffer, "FORM_A6"); break;
   case DeviceForm::FORM_A7:                    strcpy (pszBuffer, "FORM_A7"); break;
   case DeviceForm::FORM_A8:                    strcpy (pszBuffer, "FORM_A8"); break;
   case DeviceForm::FORM_A9:                    strcpy (pszBuffer, "FORM_A9"); break;
   case DeviceForm::FORM_A10:                   strcpy (pszBuffer, "FORM_A10"); break;
   case DeviceForm::FORM_B_PLUS:                strcpy (pszBuffer, "FORM_B_PLUS"); break;
   case DeviceForm::FORM_B0:                    strcpy (pszBuffer, "FORM_B0"); break;
   case DeviceForm::FORM_B1:                    strcpy (pszBuffer, "FORM_B1"); break;
   case DeviceForm::FORM_B2:                    strcpy (pszBuffer, "FORM_B2"); break;
   case DeviceForm::FORM_B3:                    strcpy (pszBuffer, "FORM_B3"); break;
   case DeviceForm::FORM_B4:                    strcpy (pszBuffer, "FORM_B4"); break;
   case DeviceForm::FORM_B5:                    strcpy (pszBuffer, "FORM_B5"); break;
   case DeviceForm::FORM_B5_EXTRA:              strcpy (pszBuffer, "FORM_B5_EXTRA"); break;
   case DeviceForm::FORM_B6:                    strcpy (pszBuffer, "FORM_B6"); break;
   case DeviceForm::FORM_B6_C4:                 strcpy (pszBuffer, "FORM_B6_C4"); break;
   case DeviceForm::FORM_B7:                    strcpy (pszBuffer, "FORM_B7"); break;
   case DeviceForm::FORM_B8:                    strcpy (pszBuffer, "FORM_B8"); break;
   case DeviceForm::FORM_B9:                    strcpy (pszBuffer, "FORM_B9"); break;
   case DeviceForm::FORM_B10:                   strcpy (pszBuffer, "FORM_B10"); break;
   case DeviceForm::FORM_JIS_B0:                strcpy (pszBuffer, "FORM_JIS_B0"); break;
   case DeviceForm::FORM_JIS_B1:                strcpy (pszBuffer, "FORM_JIS_B1"); break;
   case DeviceForm::FORM_JIS_B2:                strcpy (pszBuffer, "FORM_JIS_B2"); break;
   case DeviceForm::FORM_JIS_B3:                strcpy (pszBuffer, "FORM_JIS_B3"); break;
   case DeviceForm::FORM_JIS_B4:                strcpy (pszBuffer, "FORM_JIS_B4"); break;
   case DeviceForm::FORM_JIS_B5:                strcpy (pszBuffer, "FORM_JIS_B5"); break;
   case DeviceForm::FORM_JIS_B6:                strcpy (pszBuffer, "FORM_JIS_B6"); break;
   case DeviceForm::FORM_JIS_B7:                strcpy (pszBuffer, "FORM_JIS_B7"); break;
   case DeviceForm::FORM_JIS_B8:                strcpy (pszBuffer, "FORM_JIS_B8"); break;
   case DeviceForm::FORM_JIS_B9:                strcpy (pszBuffer, "FORM_JIS_B9"); break;
   case DeviceForm::FORM_JIS_B10:               strcpy (pszBuffer, "FORM_JIS_B10"); break;
   case DeviceForm::FORM_JIS_EXEC:              strcpy (pszBuffer, "FORM_JIS_EXEC"); break;
   case DeviceForm::FORM_C0:                    strcpy (pszBuffer, "FORM_C0"); break;
   case DeviceForm::FORM_C1:                    strcpy (pszBuffer, "FORM_C1"); break;
   case DeviceForm::FORM_C2:                    strcpy (pszBuffer, "FORM_C2"); break;
   case DeviceForm::FORM_C3:                    strcpy (pszBuffer, "FORM_C3"); break;
   case DeviceForm::FORM_C4:                    strcpy (pszBuffer, "FORM_C4"); break;
   case DeviceForm::FORM_C5:                    strcpy (pszBuffer, "FORM_C5"); break;
   case DeviceForm::FORM_C6:                    strcpy (pszBuffer, "FORM_C6"); break;
   case DeviceForm::FORM_C6_C5:                 strcpy (pszBuffer, "FORM_C6_C5"); break;
   case DeviceForm::FORM_C7:                    strcpy (pszBuffer, "FORM_C7"); break;
   case DeviceForm::FORM_C7_C6:                 strcpy (pszBuffer, "FORM_C7_C6"); break;
   case DeviceForm::FORM_C8:                    strcpy (pszBuffer, "FORM_C8"); break;
   case DeviceForm::FORM_C9:                    strcpy (pszBuffer, "FORM_C9"); break;
   case DeviceForm::FORM_C10:                   strcpy (pszBuffer, "FORM_C10"); break;
   case DeviceForm::FORM_E1:                    strcpy (pszBuffer, "FORM_E1"); break;
   case DeviceForm::FORM_RA0:                   strcpy (pszBuffer, "FORM_RA0"); break;
   case DeviceForm::FORM_RA1:                   strcpy (pszBuffer, "FORM_RA1"); break;
   case DeviceForm::FORM_RA2:                   strcpy (pszBuffer, "FORM_RA2"); break;
   case DeviceForm::FORM_SRA0:                  strcpy (pszBuffer, "FORM_SRA0"); break;
   case DeviceForm::FORM_SRA1:                  strcpy (pszBuffer, "FORM_SRA1"); break;
   case DeviceForm::FORM_SRA2:                  strcpy (pszBuffer, "FORM_SRA2"); break;
   case DeviceForm::FORM_US_FANFOLD:            strcpy (pszBuffer, "FORM_US_FANFOLD"); break;
   case DeviceForm::FORM_FANFOLD_1:             strcpy (pszBuffer, "FORM_FANFOLD_1"); break;
   case DeviceForm::FORM_FANFOLD_2:             strcpy (pszBuffer, "FORM_FANFOLD_2"); break;
   case DeviceForm::FORM_FANFOLD_3:             strcpy (pszBuffer, "FORM_FANFOLD_3"); break;
   case DeviceForm::FORM_FANFOLD_4:             strcpy (pszBuffer, "FORM_FANFOLD_4"); break;
   case DeviceForm::FORM_FANFOLD_5:             strcpy (pszBuffer, "FORM_FANFOLD_5"); break;
   case DeviceForm::FORM_GERMAN_LEGAL_FANFOLD:  strcpy (pszBuffer, "FORM_GERMAN_LEGAL_FANFOLD"); break;
   case DeviceForm::FORM_EUROPEAN_FANFOLD:      strcpy (pszBuffer, "FORM_EUROPEAN_FANFOLD"); break;
   case DeviceForm::FORM_PANORAMIC:             strcpy (pszBuffer, "FORM_PANORAMIC"); break;
   case DeviceForm::FORM_PHOTO_100_150:         strcpy (pszBuffer, "FORM_PHOTO_100_150"); break;
   case DeviceForm::FORM_PHOTO_200_300:         strcpy (pszBuffer, "FORM_PHOTO_200_300"); break;
   case DeviceForm::FORM_PHOTO_4_6:             strcpy (pszBuffer, "FORM_PHOTO_4_6"); break;
   case DeviceForm::FORM_PLOTTER_SIZE_A:        strcpy (pszBuffer, "FORM_PLOTTER_SIZE_A"); break;
   case DeviceForm::FORM_PLOTTER_SIZE_B:        strcpy (pszBuffer, "FORM_PLOTTER_SIZE_B"); break;
   case DeviceForm::FORM_PLOTTER_SIZE_C:        strcpy (pszBuffer, "FORM_PLOTTER_SIZE_C"); break;
   case DeviceForm::FORM_PLOTTER_SIZE_D:        strcpy (pszBuffer, "FORM_PLOTTER_SIZE_D"); break;
   case DeviceForm::FORM_PLOTTER_SIZE_E:        strcpy (pszBuffer, "FORM_PLOTTER_SIZE_E"); break;
   case DeviceForm::FORM_PLOTTER_SIZE_F:        strcpy (pszBuffer, "FORM_PLOTTER_SIZE_F"); break;
   case DeviceForm::FORM_US_A_ARCHITECTURAL:    strcpy (pszBuffer, "FORM_US_A_ARCHITECTURAL"); break;
   case DeviceForm::FORM_US_B_ARCHITECTURAL:    strcpy (pszBuffer, "FORM_US_B_ARCHITECTURAL"); break;
   case DeviceForm::FORM_US_C_ARCHITECTURAL:    strcpy (pszBuffer, "FORM_US_C_ARCHITECTURAL"); break;
   case DeviceForm::FORM_US_D_ARCHITECTURAL:    strcpy (pszBuffer, "FORM_US_D_ARCHITECTURAL"); break;
   case DeviceForm::FORM_US_E_ARCHITECTURAL:    strcpy (pszBuffer, "FORM_US_E_ARCHITECTURAL"); break;
   case DeviceForm::FORM_SUPER_A3_B:            strcpy (pszBuffer, "FORM_SUPER_A3_B"); break;
   case DeviceForm::FORM_SUPER_A:               strcpy (pszBuffer, "FORM_SUPER_A"); break;
   case DeviceForm::FORM_SUPER_B:               strcpy (pszBuffer, "FORM_SUPER_B"); break;
   case DeviceForm::FORM_8_5_X_12_4:            strcpy (pszBuffer, "FORM_8_5_X_12_4"); break;
   case DeviceForm::FORM_8_X_10_5:              strcpy (pszBuffer, "FORM_8_X_10_5"); break;
   case DeviceForm::FORM_8_25_X_13:             strcpy (pszBuffer, "FORM_8_25_X_13"); break;
   case DeviceForm::FORM_170_X_210:             strcpy (pszBuffer, "FORM_170_X_210"); break;
   case DeviceForm::FORM_182_X_210:             strcpy (pszBuffer, "FORM_182_X_210"); break;
   case DeviceForm::FORM_ROLL_76_2MM:           strcpy (pszBuffer, "FORM_ROLL_76_2MM"); break;
   case DeviceForm::FORM_ROLL_69_5MM:           strcpy (pszBuffer, "FORM_ROLL_69_5MM"); break;

   default:
   {
      sprintf(pszBuffer, "w%dh%d", iCx, iCy);
      break;
   }
   }
}

char *
GetResolutionName (DeviceResolution *pDR)
{
   switch(pDR->getID ())
   {
   case DeviceResolution::RESOLUTION_UNLISTED:                return "RESOLUTION_UNLISTED";
   case DeviceResolution::RESOLUTION_NONE:                    return "RESOLUTION_NONE";
   case DeviceResolution::RESOLUTION_60_X_60:                 return "RESOLUTION_60_X_60";
   case DeviceResolution::RESOLUTION_60_X_72:                 return "RESOLUTION_60_X_72";
   case DeviceResolution::RESOLUTION_60_X_180:                return "RESOLUTION_60_X_180";
   case DeviceResolution::RESOLUTION_60_X_360:                return "RESOLUTION_60_X_360";
   case DeviceResolution::RESOLUTION_72_X_72:                 return "RESOLUTION_72_X_72";
   case DeviceResolution::RESOLUTION_75_X_75:                 return "RESOLUTION_75_X_75";
   case DeviceResolution::RESOLUTION_80_X_60:                 return "RESOLUTION_80_X_60";
   case DeviceResolution::RESOLUTION_80_X_72:                 return "RESOLUTION_80_X_72";
   case DeviceResolution::RESOLUTION_90_X_180:                return "RESOLUTION_90_X_180";
   case DeviceResolution::RESOLUTION_90_X_360:                return "RESOLUTION_90_X_360";
   case DeviceResolution::RESOLUTION_90_X_60:                 return "RESOLUTION_90_X_60";
   case DeviceResolution::RESOLUTION_90_X_72:                 return "RESOLUTION_90_X_72";
   case DeviceResolution::RESOLUTION_90_X_90:                 return "RESOLUTION_90_X_90";
   case DeviceResolution::RESOLUTION_100_X_100:               return "RESOLUTION_100_X_100";
   case DeviceResolution::RESOLUTION_120_X_60:                return "RESOLUTION_120_X_60";
   case DeviceResolution::RESOLUTION_120_X_72:                return "RESOLUTION_120_X_72";
   case DeviceResolution::RESOLUTION_120_X_120:               return "RESOLUTION_120_X_120";
   case DeviceResolution::RESOLUTION_120_X_144:               return "RESOLUTION_120_X_144";
   case DeviceResolution::RESOLUTION_120_X_180:               return "RESOLUTION_120_X_180";
   case DeviceResolution::RESOLUTION_120_X_360:               return "RESOLUTION_120_X_360";
   case DeviceResolution::RESOLUTION_144_X_72:                return "RESOLUTION_144_X_72";
   case DeviceResolution::RESOLUTION_150_X_150:               return "RESOLUTION_150_X_150";
   case DeviceResolution::RESOLUTION_180_X_180:               return "RESOLUTION_180_X_180";
   case DeviceResolution::RESOLUTION_180_X_360:               return "RESOLUTION_180_X_360";
   case DeviceResolution::RESOLUTION_200_X_200:               return "RESOLUTION_200_X_200";
   case DeviceResolution::RESOLUTION_240_X_60:                return "RESOLUTION_240_X_60";
   case DeviceResolution::RESOLUTION_240_X_72:                return "RESOLUTION_240_X_72";
   case DeviceResolution::RESOLUTION_240_X_144:               return "RESOLUTION_240_X_144";
   case DeviceResolution::RESOLUTION_240_X_240:               return "RESOLUTION_240_X_240";
   case DeviceResolution::RESOLUTION_300_X_300:               return "RESOLUTION_300_X_300";
   case DeviceResolution::RESOLUTION_360_X_180:               return "RESOLUTION_360_X_180";
   case DeviceResolution::RESOLUTION_360_X_360:               return "RESOLUTION_360_X_360";
   case DeviceResolution::RESOLUTION_360_X_720:               return "RESOLUTION_360_X_720";
   case DeviceResolution::RESOLUTION_400_X_400:               return "RESOLUTION_400_X_400";
   case DeviceResolution::RESOLUTION_600_X_300:               return "RESOLUTION_600_X_300";
   case DeviceResolution::RESOLUTION_600_X_600:               return "RESOLUTION_600_X_600";
   case DeviceResolution::RESOLUTION_720_X_360:               return "RESOLUTION_720_X_360";
   case DeviceResolution::RESOLUTION_720_X_720:               return "RESOLUTION_720_X_720";
   case DeviceResolution::RESOLUTION_1200_X_600:              return "RESOLUTION_1200_X_600";
   case DeviceResolution::RESOLUTION_1200_X_1200:             return "RESOLUTION_1200_X_1200";
   case DeviceResolution::RESOLUTION_1440_X_720:              return "RESOLUTION_1440_X_720";
   case DeviceResolution::RESOLUTION_2880_X_720:              return "RESOLUTION_2880_X_720";
   case DeviceResolution::RESOLUTION_DRAFT:                   return "RESOLUTION_DRAFT";
   case DeviceResolution::RESOLUTION_FINE:                    return "RESOLUTION_FINE";
   case DeviceResolution::RESOLUTION_GROUP_3:                 return "RESOLUTION_GROUP_3";
   case DeviceResolution::RESOLUTION_GROUP_4:                 return "RESOLUTION_GROUP_4";
   case DeviceResolution::RESOLUTION_HIGH:                    return "RESOLUTION_HIGH";
   case DeviceResolution::RESOLUTION_LOW:                     return "RESOLUTION_LOW";
   case DeviceResolution::RESOLUTION_MEDIUM:                  return "RESOLUTION_MEDIUM";
   case DeviceResolution::RESOLUTION_NORMAL:                  return "RESOLUTION_NORMAL";
   case DeviceResolution::RESOLUTION_PHOTO_QUALITY:           return "RESOLUTION_PHOTO_QUALITY";
   case DeviceResolution::RESOLUTION_PRESENTATION:            return "RESOLUTION_PRESENTATION";
   }

   return "";
}

char *
quoteString (const char *pszString)
{
   if (!pszString)
      return 0;

   int         cbNewString = 0;
   const char *pszFrom     = pszString;

   while (*pszFrom)
   {
      switch (*pszFrom++)
      {
      case ' ':
      case '"':
      case '%':
         cbNewString += 3;
         break;
      default:
         cbNewString++;
         break;
      }
   }

   if (cbNewString)
   {
      cbNewString++;

      char *pszNewString = (char *)calloc (1, cbNewString);
      char *pszTo        = 0;

      if (pszNewString)
      {
         pszFrom = pszString;
         pszTo   = pszNewString;

         while (*pszFrom)
         {
            switch (*pszFrom)
            {
            case ' ': strcat (pszTo, "%20"); pszTo += strlen (pszTo); break;
            case '"': strcat (pszTo, "%22"); pszTo += strlen (pszTo); break;
            case '%': strcat (pszTo, "%25"); pszTo += strlen (pszTo); break;
            default:  *pszTo++ = *pszFrom; break;
            }
            pszFrom++;
         }

         *pszTo = '\0';

         return pszNewString;
      }
   }

   return 0;
}

int
main (int argc, char *argv[])
{
   char                achBuffer1[BUFSIZE1];
   std::ostringstream  oss;
   std::string         stringOss;
   char               *pszDriverLibrary      = 0;
   PSZ                 pszJobProperties      = 0;
   PSZRO               pszDeviceName         = 0;
   GModule            *hmodDevice            = 0;
   Device             *pDevice               = 0;
   PFNNEWDEVICEWARGS   pfnNewDeviceWArgs     = 0;
   Enumeration        *pEnum                 = 0;
   DeviceForm         *pDF                   = 0;
   DeviceMedia        *pDM                   = 0;
   DeviceTray         *pDT                   = 0;
   DeviceResolution   *pDR                   = 0;
   int                 rc                    = 0;

   Omni::initialize ();

   for (int i = 1; i < argc; i++)
   {
      if (0 == strcmp (argv[i], "--driver"))
      {
         pszDriverLibrary = argv[++i];
      }
      else if (0 == strncasecmp ("-sproperties=", argv[i], 13))
      {
         char *pszJobProp   = argv[i] + 13;
         int   iSpaceNeeded = strlen (pszJobProp) + 1;

         if (pszJobProperties)
         {
            free ((void *)pszJobProperties);
            pszJobProperties = 0;
         }

         pszJobProperties = (char *)malloc (iSpaceNeeded);
         if (pszJobProperties)
         {
            strcpy (pszJobProperties, pszJobProp);
         }
         else
         {
            std::cerr << "Error:  Out of memory" << std::endl;

            rc = __LINE__;
            goto BUGOUT;
         }

#ifndef RETAIL
         DebugOutput::applyAllDebugOutput (pszJobProperties);
#endif
      }
      else
      {
         printUsage (argv[0]);
      }
   }

   if (!pszDriverLibrary)
   {
      printUsage (argv[0]);

      rc = __LINE__;
      goto BUGOUT;
   }

   if (0 == strstr (pszDriverLibrary, ".so"))
   {
      oss << "lib"
          << pszDriverLibrary
          << ".so";

      stringOss        = oss.str ();
      pszDriverLibrary = (char *)stringOss.c_str ();
   }

   std::cerr << "Trying to load " << pszDriverLibrary << std::endl;

   if (!Omni::openDeviceLibrary (pszDriverLibrary, &hmodDevice))
   {
      rc = __LINE__;
      goto BUGOUT;
   }

   g_module_symbol (hmodDevice,
                    "newDeviceW_JopProp_Advanced",
                    (gpointer *)&pfnNewDeviceWArgs);

   if (pfnNewDeviceWArgs)
   {
      pDevice = pfnNewDeviceWArgs (pszJobProperties, true);
   }

   if (!pDevice)
   {
      std::cerr << "Error: Could not instantiate a new device with \"" << pszJobProperties << "\"" << std::endl;

      rc = __LINE__;
      goto BUGOUT;
   }

   if (pDevice->hasError ())
   {
      std::cerr << "Error: The device had an error with \"" << pszJobProperties << "\"" << std::endl;

      rc = __LINE__;
      goto BUGOUT;
   }

   pszDeviceName = pDevice->getDeviceName ();

   std::cout << "*PPD-Adobe: \"4.3\"" << std::endl;
   std::cout << "*%PPD file for CUPS/Omni." << std::endl;
   std::cout << "*%This PPD file may be freely used and distributed under the terms of" << std::endl;
   std::cout << "*%the GNU LGPL." << std::endl;
   std::cout << "*FormatVersion:	\"4.3\"" << std::endl;
   std::cout << "*FileVersion:	\"" << pDevice->getVersion () << "\"" << std::endl;
   std::cout << "*LanguageVersion: English" << std::endl;
   std::cout << "*LanguageEncoding: ISOLatin1" << std::endl;
   std::cout << "*PCFileName:	\"" << pDevice->getShortName () << ".ppd\"" << std::endl;
   std::cout << "*Manufacturer:	\"" << pDevice->getDriverName () << "\"" << std::endl;
   std::cout << "*Product:	\"(Omni v" << pDevice->getVersion () << ")\"" << std::endl;
   std::cout << "*ModelName:	\"" << pDevice->getDeviceName () << "\"" << std::endl;
   std::cout << "*NickName:	\"" << pszDeviceName << ", Foomatic + omni\"" << std::endl;
   {
      std::ostringstream oss;

      oss << pDevice->getLibraryName ();
#if 0
      if (  pszJobProperties
         && *pszJobProperties
         )
      {
         oss << " " << pszJobProperties;
      }
#else
      std::string *pstringJP = pDevice->queryCurrentJobProperties ();
      if (pstringJP)
      {
         oss << " " << *pstringJP;
         delete pstringJP;
      }
#endif

      char *pszQuoted = quoteString (oss.str ().c_str ());

      if (pszQuoted)
      {
         std::cout << "*ShortNickName:	\"" << pszQuoted << "\"" << std::endl;

         free (pszQuoted);
      }
   }
   std::cout << "*PSVersion:	\"(3010.000) 550\"" << std::endl;
   std::cout << "*LanguageLevel:	\"3\"" << std::endl;

   if (pDevice->hasCapability (Capability::COLOR))
   {
      std::cout << "*ColorDevice:	True" << std::endl;
      std::cout << "*DefaultColorSpace:	RGB" << std::endl;
   }
   else
   {
      std::cout << "*ColorDevice:	False" << std::endl;
      std::cout << "*DefaultColorSpace:	Gray" << std::endl;
   }

   std::cout << "*FileSystem:	False" << std::endl;
   std::cout << "*LandscapeOrientation: Plus90" << std::endl;
   std::cout << "*TTRasterizer:	Type42" << std::endl;

   std::cout << "*cupsVersion:	1.1" << std::endl;
// *cupsModelNumber @TBD
   std::cout << "*cupsManualCopies: True" << std::endl;
   std::cout << "*cupsFilter:	\"application/vnd.cups-raster 100 CUPSToOmni\"" << std::endl;

   std::cout << std::endl;

   std::cout << "*OpenUI *PageSize/Media Size: PickOne" << std::endl;
   std::cout << "*OrderDependency: 10 AnySetup *PageSize" << std::endl;

   pDF   = pDevice->getCurrentForm ();
   pEnum = pDF->getEnumeration ();

   std::cout << "*DefaultPageSize: ";

   memset (achBuffer1, 0, sizeof (achBuffer1));
   writePostscriptName (pDF, achBuffer1);
   std::cout << achBuffer1 << std::endl;

   while (pEnum->hasMoreElements ())
   {

      pDF = (DeviceForm *)pEnum->nextElement ();

      HardCopyCap *pHCC = pDF->getHardCopyCap ();
      int          iCx;
      int          iCy;

      iCx = (int)(72.0 * (float)pHCC->getCx () / TMM_TO_INCHES);
      iCy = (int)(72.0 * (float)pHCC->getCy () / TMM_TO_INCHES);

      std::cout << "*PageSize ";

      memset (achBuffer1, 0, sizeof (achBuffer1));
      writePostscriptName (pDF, achBuffer1);
      std::cout << achBuffer1;

      std::cout << "/"
                << pDF->getName ()
                << ":\t\"<</PageSize["
                << iCx
                << " "
                << iCy
                << "]/ImagingBBox null>>setpagedevice\""
                << std::endl;

      delete pDF;
   }

   std::cout << "*CloseUI: *PageSize" << std::endl;

   delete pEnum;

   std::cout << std::endl;

   std::cout << "*OpenUI *PageRegion: PickOne" << std::endl;
   std::cout << "*OrderDependency: 10 AnySetup *PageRegion" << std::endl;

   pDF   = pDevice->getCurrentForm ();
   pEnum = pDF->getEnumeration ();

   std::cout << "*DefaultPageRegion: ";

   memset (achBuffer1, 0, sizeof (achBuffer1));
   writePostscriptName (pDF, achBuffer1);
   std::cout << achBuffer1 << std::endl;

   while (pEnum->hasMoreElements ())
   {
      pDF = (DeviceForm *)pEnum->nextElement ();

      HardCopyCap *pHCC = pDF->getHardCopyCap ();
      int          iCx;
      int          iCy;

      iCx = (int)(72.0 * (float)pHCC->getCx () / TMM_TO_INCHES);
      iCy = (int)(72.0 * (float)pHCC->getCy () / TMM_TO_INCHES);

      std::cout << "*PageRegion ";

      memset(achBuffer1, 0, sizeof (achBuffer1));
      writePostscriptName (pDF, achBuffer1);
      std::cout << achBuffer1;

      std::cout << "/"
                << pDF->getName ()
                << ":\t\"<</PageRegion["
                << iCx
                << " "
                << iCy
                << "]/ImagingBBox null>>setpagedevice\""
                << std::endl;

      delete pDF;
   }

   std::cout << "*CloseUI: *PageRegion" << std::endl;

   delete pEnum;

   std::cout << std::endl;

   pDF   = pDevice->getCurrentForm ();
   pEnum = pDF->getEnumeration ();

   std::cout << "*DefaultImageableArea: ";

   memset (achBuffer1, 0, sizeof (achBuffer1));
   writePostscriptName (pDF, achBuffer1);
   std::cout << achBuffer1 << std::endl;

   while (pEnum->hasMoreElements ())
   {
      pDF = (DeviceForm *)pEnum->nextElement ();

      HardCopyCap *pHCC    = pDF->getHardCopyCap ();
      float        iCx     = 72.0 * (float)pHCC->getCx         () / TMM_TO_INCHES;
      float        iCy     = 72.0 * (float)pHCC->getCy         () / TMM_TO_INCHES;
      float        iLeft   = 72.0 * (float)pHCC->getLeftClip   () / TMM_TO_INCHES;
      float        iRight  = 72.0 * (float)pHCC->getRightClip  () / TMM_TO_INCHES;
      float        iTop    = 72.0 * (float)pHCC->getTopClip    () / TMM_TO_INCHES;
      float        iBottom = 72.0 * (float)pHCC->getBottomClip () / TMM_TO_INCHES;

      std::cout << "*ImageableArea ";

      memset (achBuffer1, 0, sizeof (achBuffer1));
      writePostscriptName (pDF, achBuffer1);
      std::cout << achBuffer1;

      std::cout << "/"
                << pDF->getName ()
                << ":\t\""
                << iLeft
                << " "
                << iBottom
                << " "
                << iCx - iRight
                << " "
                << iCy - iTop
                << "\""
                << std::endl;

      delete pDF;
   }

   delete pEnum;

   std::cout << std::endl;

   pDF   = pDevice->getCurrentForm ();
   pEnum = pDF->getEnumeration ();

   std::cout << "*DefaultPaperDimension: ";

   memset(achBuffer1, 0, sizeof (achBuffer1));
   writePostscriptName (pDF, achBuffer1);
   std::cout << achBuffer1 << std::endl;

   while (pEnum->hasMoreElements ())
   {
      pDF = (DeviceForm *)pEnum->nextElement ();

      HardCopyCap *pHCC    = pDF->getHardCopyCap ();
      int          iCx     = (int)(72.0 * (float)pHCC->getCx () / TMM_TO_INCHES);
      int          iCy     = (int)(72.0 * (float)pHCC->getCy () / TMM_TO_INCHES);

      std::cout << "*PaperDimension ";

      memset (achBuffer1, 0, sizeof (achBuffer1));
      writePostscriptName (pDF, achBuffer1);
      std::cout << achBuffer1;

      std::cout << "/"
                << pDF->getName ()
                << ":\t\""
                << iCx
                << " "
                << iCy
                << "\""
                << std::endl;

      delete pDF;
   }

   delete pEnum;

   std::cout << std::endl;

   std::cout << "*OpenUI *ColorModel: PickOne" << std::endl;
   std::cout << "*OrderDependency: 10 AnySetup *ColorModel" << std::endl;

   if (pDevice->hasCapability (Capability::COLOR))
   {
      std::cout << "*DefaultColorModel: RGB" << std::endl;
      std::cout << "*ColorModel Gray/Grayscale:\t\"<</cupsColorSpace "
                << (unsigned long int)CUPS_CSPACE_W
                << "/cupsColorOrder "
                << (unsigned long int)CUPS_ORDER_CHUNKED
                << "/cupsBitsPerColor 1>>setpagedevice\""
                << std::endl;
      std::cout << "*ColorModel RGB/Color:\t\"<</cupsColorSpace "
                << (unsigned long int)CUPS_CSPACE_RGB
                << "/cupsColorOrder "
                << (unsigned long int)CUPS_ORDER_CHUNKED
                << "/cupsBitsPerColor 8>>setpagedevice\"" // @TBD - shouldn't this be 24 bits per pel?
                << std::endl;
   }
   else
   {
      std::cout << "*DefaultColorModel: Gray" << std::endl;
      std::cout << "*ColorModel Gray/Grayscale:\t\"<</cupsColorSpace "
                << (unsigned long int)CUPS_CSPACE_W
                << "/cupsColorOrder "
                << (unsigned long int)CUPS_ORDER_CHUNKED
                << "/cupsBitsPerColor 1>>setpagedevice\""
                << std::endl;
   }

   std::cout << "*CloseUI: *ColorModel" << std::endl;

   std::cout << std::endl;

   std::cout << "*OpenUI *MediaType: PickOne" << std::endl;
   std::cout << "*OrderDependency: 10 AnySetup *MediaType" << std::endl;

   pDM = pDevice->getCurrentMedia ();

   std::cout << "*DefaultMediaType: " << DeviceMedia::IDToName (pDM->getID ()) << std::endl;

   pEnum = pDM->getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      pDM = (DeviceMedia *)pEnum->nextElement ();

      std::cout << "*MediaType "
                << DeviceMedia::IDToName (pDM->getID ())
                << "/"
                << pDM->getName ()
                << ":\t\"<</MediaType("
                << pDM->getName ()        // @TBD - should this be our settable name?
                << ")>>setpagedevice\""
                << std::endl;

      delete pDM;
   }

   std::cout << "*CloseUI: *MediaType" << std::endl;

   delete pEnum;

   std::cout << std::endl;

   std::cout << "*OpenUI *InputSlot: PickOne" << std::endl;
   std::cout << "*OrderDependency: 10 AnySetup *InputSlot" << std::endl;

   pDT = pDevice->getCurrentTray ();

   std::cout << "*DefaultInputSlot: " << DeviceTray::IDToName (pDT->getID ()) << std::endl;

   pEnum = pDT->getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      pDT = (DeviceTray *)pEnum->nextElement ();

      std::cout << "*InputSlot "
                << DeviceTray::IDToName (pDT->getID ())
                << "/"
                << pDT->getName ()
                << ":\t\"<</MediaClass("
                << pDT->getName ()        // @TBD - should this be our settable name?
                << ")>>setpagedevice\""
                << std::endl;

      delete pDT;
   }

   std::cout << "*CloseUI: *InputSlot" << std::endl;

   delete pEnum;

   std::cout << std::endl;

   std::cout << "*OpenUI *Dither: PickOne" << std::endl;
   std::cout << "*OrderDependency: 10 AnySetup *Dither" << std::endl;

   pEnum = DeviceDither::getEnumeration ();

   std::cout << "*DefaultDither: " << pDevice->getCurrentDitherID () << std::endl;

   while (pEnum->hasMoreElements ())
   {
      char *pszID = (char *)pEnum->nextElement ();

      std::cout << "*Dither "
                << pszID
                << "/"
                << DeviceDither::getName (pDevice,pszID)
                << ":\t\"<</OutputType("
                // @TBD - should this be our settable name?
                << DeviceDither::getName (pDevice,pszID)
                << ")>>setpagedevice\""
                << std::endl;
   }

   std::cout << "*CloseUI: *Dither" << std::endl;

   delete pEnum;

   std::cout << std::endl;

   std::cout << "*OpenUI *Resolution: PickOne" << std::endl;
   std::cout << "*OrderDependency: 20 AnySetup *Resolution" << std::endl;

   pDR = pDevice->getCurrentResolution ();

///std::cout << "*DefaultResolution: " << DeviceResolution::IDToName (pDR->getID ()) << std::endl;

   std::cout << "*DefaultResolution: " << GetResolutionName(pDR) << std::endl;

   pEnum = pDR->getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      pDR = (DeviceResolution *)pEnum->nextElement ();

      std::cout << "*Resolution "
////////////////<< DeviceResolution::IDToName (pDR->getID ())
                << GetResolutionName(pDR)
                << "/"
                << pDR->getName ()
                << ":\t\"<</HWResolution["
                << pDR->getXRes ()
                << " "
                << pDR->getYRes ()
                << "]/cupsCompression "
                << pDR->getID ()        // @TBD - correct?
                << ">>setpagedevice\""
                << std::endl;

      delete pDR;
   }

   std::cout << "*CloseUI: *Resolution" << std::endl;

   delete pEnum;

   std::cout << std::endl;

   std::cout << "*DefaultFont: Courier" << std::endl;
   std::cout << "*Font AvantGarde-Book: Standard \"(001.006S)\" Standard ROM" << std::endl;
   std::cout << "*Font AvantGarde-BookOblique: Standard \"(001.006S)\" Standard ROM" << std::endl;
   std::cout << "*Font AvantGarde-Demi: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font AvantGarde-DemiOblique: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font Bookman-Demi: Standard \"(001.004S)\" Standard ROM" << std::endl;
   std::cout << "*Font Bookman-DemiItalic: Standard \"(001.004S)\" Standard ROM" << std::endl;
   std::cout << "*Font Bookman-Light: Standard \"(001.004S)\" Standard ROM" << std::endl;
   std::cout << "*Font Bookman-LightItalic: Standard \"(001.004S)\" Standard ROM" << std::endl;
   std::cout << "*Font Courier: Standard \"(002.004S)\" Standard ROM" << std::endl;
   std::cout << "*Font Courier-Bold: Standard \"(002.004S)\" Standard ROM" << std::endl;
   std::cout << "*Font Courier-BoldOblique: Standard \"(002.004S)\" Standard ROM" << std::endl;
   std::cout << "*Font Courier-Oblique: Standard \"(002.004S)\" Standard ROM" << std::endl;
   std::cout << "*Font Helvetica: Standard \"(001.006S)\" Standard ROM" << std::endl;
   std::cout << "*Font Helvetica-Bold: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font Helvetica-BoldOblique: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font Helvetica-Narrow: Standard \"(001.006S)\" Standard ROM" << std::endl;
   std::cout << "*Font Helvetica-Narrow-Bold: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font Helvetica-Narrow-BoldOblique: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font Helvetica-Narrow-Oblique: Standard \"(001.006S)\" Standard ROM" << std::endl;
   std::cout << "*Font Helvetica-Oblique: Standard \"(001.006S)\" Standard ROM" << std::endl;
   std::cout << "*Font NewCenturySchlbk-Bold: Standard \"(001.009S)\" Standard ROM" << std::endl;
   std::cout << "*Font NewCenturySchlbk-BoldItalic: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font NewCenturySchlbk-Italic: Standard \"(001.006S)\" Standard ROM" << std::endl;
   std::cout << "*Font NewCenturySchlbk-Roman: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font Palatino-Bold: Standard \"(001.005S)\" Standard ROM" << std::endl;
   std::cout << "*Font Palatino-BoldItalic: Standard \"(001.005S)\" Standard ROM" << std::endl;
   std::cout << "*Font Palatino-Italic: Standard \"(001.005S)\" Standard ROM" << std::endl;
   std::cout << "*Font Palatino-Roman: Standard \"(001.005S)\" Standard ROM" << std::endl;
   std::cout << "*Font Symbol: Special \"(001.007S)\" Special ROM" << std::endl;
   std::cout << "*Font Times-Bold: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font Times-BoldItalic: Standard \"(001.009S)\" Standard ROM" << std::endl;
   std::cout << "*Font Times-Italic: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font Times-Roman: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font ZapfChancery-MediumItalic: Standard \"(001.007S)\" Standard ROM" << std::endl;
   std::cout << "*Font ZapfDingbats: Special \"(001.004S)\" Standard ROM" << std::endl;

   std::cout << "*%%End of " << pszDeviceName << ".ppd" << std::endl;

BUGOUT:
   delete pDevice; pDevice = 0;

   if (hmodDevice)
   {
      g_module_close (hmodDevice);
      hmodDevice = 0;
   }

   if (pszJobProperties)
   {
      free ((void *)pszJobProperties);
      pszJobProperties = 0;
   }

   Omni::terminate ();

   return rc;
}
