/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2002
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Device.hpp"
#include "Omni.hpp"

Device *
loadDevice (PSZCRO pszFullDeviceName,
            PSZCRO pszJobProperties)
{
   Device            *pDevice           = 0;
   PFNNEWDEVICEWARGS  pfnNewDeviceWArgs = 0;
   GModule           *hmodDevice        = 0;

   std::cout << "Trying to load \""
             << (pszFullDeviceName ? pszFullDeviceName : "")
             << "\" with \""
             << (pszJobProperties ? pszJobProperties : "")
             << "\""
             << std::endl;

   hmodDevice = g_module_open (pszFullDeviceName, (GModuleFlags)0);

   if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << __FUNCTION__ << ": hmodDevice = " << std::hex << hmodDevice << std::endl;

   if (!hmodDevice)
   {
      DebugOutput::getErrorStream () << "g_module_error returns " << g_module_error () << std::endl;
      return 0;
   }

   g_module_symbol (hmodDevice,
                    "newDeviceW_JopProp_Advanced",
                    (gpointer *)&pfnNewDeviceWArgs);

   if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << __FUNCTION__ << ": pfnNewDeviceWArgs = 0x" << std::hex << (int)pfnNewDeviceWArgs << std::dec << std::endl;

   if (!pfnNewDeviceWArgs)
      return 0;

   pDevice = pfnNewDeviceWArgs (pszJobProperties, true);

   return pDevice;
}

int
handleDevice (PSZCRO pszFullDeviceName,
              PSZCRO pszJobProperties)
{
   Device *pDevice = 0;

   if (strstr (pszFullDeviceName, "/lib"))
   {
//    pszFullDeviceName = strstr (pszFullDeviceName, "/lib") + 1;
   }

//#define PDC 1
#ifdef PDC
   pDevice = new OmniPDCProxy (0,                 // client exe to spawn
                               pszFullDeviceName, // device name
                               pszJobProperties,  // job properties
                               true);             // is renderer
#else
   pDevice = loadDevice (pszFullDeviceName, pszJobProperties);
#endif

   if (!pDevice)
   {
      DebugOutput::getErrorStream () << "DeviceTester4::" << __FUNCTION__ << ": the new device failed to open!" << std::endl;

      return 2;
   }

   if (pDevice->hasError ())
   {
      DebugOutput::getErrorStream () << "DeviceTester4::" << __FUNCTION__ << ": the new device has an error!" << std::endl;

      return 3;
   }

   Enumeration *pEnumKeys   = pDevice->listJobPropertyKeys ();
   Enumeration *pEnumValues = 0;

   while (pEnumKeys->hasMoreElements ())
   {
      char *pszKey = (char *)pEnumKeys->nextElement ();

      if (!pszKey)
      {
         // @TBD
         continue;
      }

      std::string *pValue = pDevice->getJobProperty (pszKey);

      if (!pValue)
      {
         // @TBD
         continue;
      }

      std::cout << pszKey << " = " << *pValue << std::endl;

      pEnumValues = pDevice->listKeyValues (pszKey);

      while (pEnumValues->hasMoreElements ())
      {
         char *pszValue = (char *)pEnumValues->nextElement ();

         std::cout << "\t"
                   << pszValue;

         std::string *pRet = pDevice->translateKeyValue (pszKey, pszValue);

         if (pRet)
         {
            std::cout << "("
                      << *pRet
                      << ")";
         }

         std::cout << std::endl;

         delete pRet;
      }

      delete pValue;
      delete pEnumValues;
   }

   delete pEnumKeys;

   pEnumKeys = pDevice->listDeviceJobPropertyKeys ();

   while (pEnumKeys->hasMoreElements ())
   {
      char *pszKey = (char *)pEnumKeys->nextElement ();

      std::cout << pszKey << " = ";

      std::string *pValue = pDevice->getJobProperty (pszKey);

      std::cout << *pValue << std::endl;

      pEnumValues = pDevice->listKeyValues (pszKey);

      while (pEnumValues->hasMoreElements ())
      {
         char *pszValue = (char *)pEnumValues->nextElement ();

         std::cout << "\t"
                   << pszValue;

         std::string *pRet = pDevice->translateKeyValue (pszKey, pszValue);

         if (pRet)
         {
            std::cout << "("
                      << *pRet
                      << ")";
         }

         std::cout << std::endl;

         delete pRet;
      }

      delete pValue;
      delete pEnumValues;
   }

   delete pEnumKeys;

   delete pDevice;

   return 0;
}

int
main (int argc, char *argv[])
{
   char *pszJobProperties  = 0;
   int   iFailed           = 0;

   if (!g_module_supported ())
   {
      std::cerr << "This program needs glib's module routines!" << std::endl;

      return 1;
   }

   if (2 > argc)
   {
      std::cerr << "Usage: " << argv[0] << " printer_name" << std::endl;

      return 1;
   }

   Omni::initialize ();

   for (int i = 1; i < argc; i++)
   {
      if (0 == strcasecmp (argv[i], "--all"))
      {
         Enumeration *pEnum = Omni::listDevices ();

         while (pEnum->hasMoreElements ())
         {
            OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

            if (pOD)
            {
               PSZCRO pszLibName = pOD->getLibraryName ();

               std::cout << "Checking: " << pszLibName << std::endl;

               iFailed |= handleDevice (pszLibName, pOD->getJobProperties ());

               delete pOD;
            }
            else
            {
               std::cerr << "Error" << std::endl;

               iFailed = 1;
            }
         }

         delete pEnum;
      }
      else if (0 == strcasecmp (argv[i], "--driver"))
      {
         std::ostringstream  oss;
         std::string         stringOss;
         char               *pszDriverLibrary = argv[i + 1];

         if (0 != strncasecmp (argv[i + 1], "lib", 3))
         {
            oss << "lib"
                << pszDriverLibrary
                << ".so";

            stringOss        = oss.str ();
            pszDriverLibrary = (char *)stringOss.c_str ();
         }

         iFailed |= handleDevice (pszDriverLibrary, pszJobProperties);
      }
      else if (0 == strncasecmp ("-sproperties=", argv[i], 13))
      {
         char *pszJobProp   = argv[i] + 13;
         int   iSpaceNeeded = strlen (pszJobProp) + 1;

         if (pszJobProperties)
         {
            free (pszJobProperties);
            pszJobProperties = 0;
         }

         pszJobProperties = (char *)malloc (iSpaceNeeded);
         if (pszJobProperties)
         {
            strcpy (pszJobProperties, pszJobProp);
         }
         else
         {
            std::cerr << "Error:  Out of memory" << std::endl;

            goto BUGOUT;
         }
      }
   }

BUGOUT:
   if (pszJobProperties)
   {
      free (pszJobProperties);
   }

   Omni::terminate ();

   return iFailed;
}
