/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef _PrintDevice
#define _PrintDevice

#include "Device.hpp"

#include <map>

#include <glib.h>
#include <gmodule.h>

class PrintDevice : public Device
{
protected:
                              PrintDevice               (PSZRO                pszDriverName,
                                                         PSZRO                pszDeviceName,
                                                         PSZRO                pszShortName,
                                                         PSZRO                pszLibraryName,
                                                         EOMNICLASS           eOmniClass,
                                                         PSZRO                pszJobProperties);

public:
   virtual                   ~PrintDevice               ();

   bool                       hasError                  ();

   void                       initialize                ();

   PSZCRO                     getVersion                ();

   PSZCRO                     getDriverName             ();
   PSZCRO                     getDeviceName             ();
   PSZCRO                     getShortName              ();
   PSZCRO                     getLibraryName            ();
   EOMNICLASS                 getOmniClass              ();

   virtual DeviceInstance    *getDeviceInstance         ();
   virtual DeviceBlitter     *getDeviceBlitter          ();

   DeviceBooklet             *getCurrentBooklet         ();
   DeviceCollation           *getCurrentCollation       ();
   DeviceCopies              *getCurrentCopies          ();
   DeviceDestination         *getCurrentDestination     ();
   PSZCRO                     getCurrentDitherID        ();
   DeviceDither              *getCurrentDither          ();
   DeviceDuplex              *getCurrentDuplex          ();
   DeviceForm                *getCurrentForm            ();
   DeviceJogging             *getCurrentJogging         ();
   DeviceMedia               *getCurrentMedia           ();
   DeviceNUp                 *getCurrentNUp             ();
   DeviceOrientation         *getCurrentOrientation     ();
   DevicePrintMode           *getCurrentPrintMode       ();
   DeviceResolution          *getCurrentResolution      ();
   DeviceScaling             *getCurrentScaling         ();
   DeviceStapling            *getCurrentStapling        ();
   DeviceTray                *getCurrentTray            ();

   virtual DeviceGamma       *getCurrentGamma           () = 0;

   DeviceCommand             *getCommands               ();
   DeviceData                *getDeviceData             ();
   DeviceString              *getDeviceString           ();

   int                        getPDLLevel               ();
   int                        getPDLSubLevel            ();
   int                        getPDLMajorRevisionLevel  ();
   int                        getPDLMinorRevisionLevel  ();

   bool                       hasCapability             (long                 lMask);
   bool                       hasRasterCapability       (long                 lMask);
   bool                       hasDeviceOption           (PSZCRO               pszDeviceOption);

   int                        getScanlineMultiple       ();

   bool                       beginJob                  ();
   bool                       beginJob                  (PSZCRO               pszJobProperties);
   bool                       newFrame                  ();
   bool                       newFrame                  (PSZCRO               pszJobProperties);
   bool                       endJob                    ();
   bool                       abortJob                  ();

   bool                       rasterize                 (PBYTE                pbBits,
                                                         PBITMAPINFO2         pbmi,
                                                         PRECTL               prectlPageLocation,
                                                         BITBLT_TYPE          eType);

   void                       setOutputStream           (FILE                *pFile);
   void                       setOutputFunction         (PFNOUTPUTFUNCTION    pfnOutputFunction,
                                                         void                *pMagicCookie);
   void                       setErrorStream            (FILE                *pFile);

   bool                       setLanguage               (int                  iLanguageID);
   int                        getLanguage               ();
   Enumeration               *getLanguages              ();
   StringResource            *getLanguageResource       ();

   bool                       sendBinaryDataToDevice    (BinaryData          *pData);
   bool                       sendBinaryDataToDevice    (DeviceForm          *pForm);
   bool                       sendBinaryDataToDevice    (DeviceTray          *pTray);
   bool                       sendBinaryDataToDevice    (DeviceMedia         *pMedia);
   bool                       sendBinaryDataToDevice    (DeviceResolution    *pResolution);
   bool                       sendBinaryDataToDevice    (PBYTE                pbData,
                                                         int                  iLength);

   bool                       sendPrintfToDevice        (BinaryData          *pData,
                                                                              ...);
   bool                       sendVPrintfToDevice       (BinaryData          *pData,
                                                         va_list              list);

   std::string               *queryDeviceJobProperties  ();

   std::string               *queryCurrentJobProperties ();
   Enumeration               *listJobPropertyKeys       ();
   Enumeration               *listDeviceJobPropertyKeys ();
   Enumeration               *listKeyValues             (PSZCRO               pszKey);
   std::string               *getJobPropertyType        (PSZCRO               pszKey);
   std::string               *getJobProperty            (PSZCRO               pszKey);

   std::string               *translateKeyValue         (PSZCRO               pszKey,
                                                         PSZCRO               pszValue);

   void                       loadLibrary               (PSZCRO               pszLibraryName);
   void                      *dlsym                     (PSZCRO               pszLibraryName,
                                                         PSZCRO               pszSymbol);

   virtual std::string        toString                  (std::ostringstream&  oss);
   friend std::ostream&       operator<<                (std::ostream&        os,
                                                         const PrintDevice&   self);

protected:
   void                       setDeviceInstance         (DeviceInstance      *pInstance);
   void                       setDeviceBlitter          (DeviceBlitter       *pBlitter);
   void                       setPDL                    (PDL                 *pPDL);

   void                       setCapabilities           (long                 lCapabilities);
   void                       setRasterCapabilities     (long                 lCapabilities);

   virtual DeviceBooklet     *getDefaultBooklet         () = 0;
   virtual DeviceCollation   *getDefaultCollation       () = 0;
   virtual DeviceCopies      *getDefaultCopies          () = 0;
   virtual DeviceDestination *getDefaultDestination     () = 0;
   virtual PSZCRO             getDefaultDitherID        () = 0;
   virtual DeviceDuplex      *getDefaultDuplex          () = 0;
   virtual DeviceForm        *getDefaultForm            () = 0;
   virtual DeviceJogging     *getDefaultJogging         () = 0;
   virtual DeviceMedia       *getDefaultMedia           () = 0;
   virtual DeviceNUp         *getDefaultNUp             () = 0;
   virtual DeviceOrientation *getDefaultOrientation     () = 0;
   virtual DevicePrintMode   *getDefaultPrintMode       () = 0;
   virtual DeviceResolution  *getDefaultResolution      () = 0;
   virtual DeviceScaling     *getDefaultScaling         () = 0;
   virtual DeviceStapling    *getDefaultStapling        () = 0;
   virtual DeviceTray        *getDefaultTray            () = 0;

   virtual DeviceCommand     *getDefaultCommands        () = 0;
   virtual DeviceData        *getDefaultData            () = 0;
   virtual DeviceString      *getDefaultString          () = 0;

private:
   void                       initializeJobProperties   ();
   void                       cleanupProperties         ();

   std::ostream        *outputStream_d;
   std::streambuf      *outputStreamBuf_d;
   bool                 fShouldDeleteOutputStream_d;

   PFNOUTPUTFUNCTION    pfnOutputFunction_d;
   void                *pMagicCookie_d;

   int                  iLanguageID_d;
   StringResource      *pLanguage_d;

   PSZRO                pszDriverName_d;
   PSZRO                pszDeviceName_d;
   PSZRO                pszShortName_d;
   PSZRO                pszLibraryName_d;
   EOMNICLASS           eOmniClass_d;
   char                *pszJobProperties_d;

   long                 lCapabilities_d;
   long                 lRasterCapabilities_d;

   DeviceInstance      *pInstance_d;
   DeviceBlitter       *pBlitter_d;
   PDL                 *pPDL_d;

   DeviceOrientation   *pOrientation_d;
   char                *pszDitherID_d;
   DeviceForm          *pForm_d;
   DeviceTray          *pTray_d;
   DeviceMedia         *pMedia_d;
   DeviceResolution    *pResolution_d;
   DevicePrintMode     *pPrintMode_d;
   DeviceCopies        *pCopies_d;
   DeviceCollation     *pCollation_d;
   DeviceDestination   *pDestination_d;
   DeviceStapling      *pStapling_d;
   DeviceJogging       *pJogging_d;
   DeviceDuplex        *pDuplex_d;
   DeviceBooklet       *pBooklet_d;
   DeviceNUp           *pNUp_d;
   DeviceScaling       *pScaling_d;

   DeviceCommand       *pCommand_d;
   DeviceData          *pData_d;
   DeviceString        *pString_d;

   char                *pszLoadedLibrary_d;  // @TBD expand to many
   GModule             *hModLibrary_d;
};

#endif
